header {* Preliminary lemmas *}

theory Lemmas imports ListPre
begin

section{* List insert*}


lemma set_list_insert[rule_format]: "(x\<in>set xs)\<longrightarrow> (\<exists> n ys. (n\<le>length ys \<and>xs = list_insert ys n x))"
apply (induct_tac xs);
apply auto;
apply (rule_tac x="0" in exI)
apply (rule_tac x=list in exI)
apply (case_tac list)
apply simp+;
apply (rule_tac x="n+1" in exI)
apply (rule_tac x="x#ys" in exI)
apply (case_tac ys)
apply simp+;
apply (rule_tac x="n+1" in exI)
apply (rule_tac x="a#ys" in exI)
apply (case_tac ys)
apply simp+;
done;

lemma  set_list_insertI[rule_format]: " \<forall> n. (x\<in>set xs)\<longrightarrow> (x\<in> set (list_insert xs n y))"
apply (induct_tac xs);
apply force;
apply clarify;
apply (case_tac n)
apply auto;
done;

lemma  set_list_insert_simp[rule_format,simp]: " \<forall> n. set (list_insert xs n y) = insert y (set xs)"
apply (induct_tac xs);
apply force;
apply clarify;
apply (case_tac n)
apply auto;
done;

lemma insert_distinct_simp[rule_format,simp]: " \<forall> n. (distinct (list_insert xs n a) = (distinct xs\<and>a\<notin>set xs))"
apply (induct_tac xs);
apply force;
apply clarify;
apply (case_tac n);
apply force;
apply (drule_tac x="nat" in spec);
apply auto;
done;


lemma filter_nth[rule_format]: "\<forall> j<length [Z\<leftarrow>ZL . P Z].\<exists> i<length ZL. [Z\<leftarrow>ZL . P Z] ! j = ZL!i"
apply (induct_tac ZL)
apply force
apply clarify
apply auto
apply (case_tac j)
apply (rule_tac x=0 in exI)
apply simp+
apply (drule_tac x=nat in spec)
apply clarify
apply (rule_tac x="Suc i" in exI)
apply force
apply (drule_tac x=j in spec)
apply clarify
apply (rule_tac x="Suc i" in exI)
apply force
done

section{*List filter*}
lemma filter_nth_double[rule_format]: "\<forall> j<length [Z\<leftarrow>ZL . P Z]. \<forall> k <length [Z\<leftarrow>ZL . P Z]. j\<noteq>k \<longrightarrow>
(\<exists> i<length ZL. \<exists> l<length ZL. i\<noteq>l\<and> [Z\<leftarrow>ZL . P Z] ! j = ZL!i\<and> [Z\<leftarrow>ZL . P Z] ! k = ZL!l)"
apply (induct_tac ZL)
apply force
apply clarify
apply auto
apply (case_tac j)
apply (rule_tac x=0 in exI)
apply simp+
apply (case_tac k)
apply force

apply simp
apply (drule_tac filter_nth)
apply clarify
apply (rule_tac x="Suc i" in exI)
apply force

apply (case_tac k)
apply simp
apply (drule_tac filter_nth)
apply clarify
apply (rule_tac x="Suc i" in exI)
apply force

apply clarify
apply (drule_tac x=nat in spec)
apply clarify
apply (drule_tac x=nata in spec)
apply clarify
apply force

apply (drule_tac x=j in spec)
apply clarify
apply (drule_tac x=k in spec)
apply force
done


lemma filter_nth_rev[rule_format]: "\<forall> j. (P (ZL!j)\<longrightarrow>j<length ZL \<longrightarrow>(\<exists> i<length [Z\<leftarrow>ZL . P Z]. ZL!j = [Z\<leftarrow>ZL . P Z]!i))"
apply (induct_tac ZL)
apply force
apply clarify
apply auto
apply (case_tac j)
apply (rule_tac x=0 in exI)
apply simp+
apply (drule_tac x=nat in spec)
apply clarify
apply (rule_tac x="Suc i" in exI)
apply force
apply (case_tac j)
apply simp+
done



section{*set*}

lemma set_exclude: "\<lbrakk>a\<notin>B;B\<subseteq>C\<rbrakk> \<Longrightarrow> B \<subseteq> C-{a}"
apply auto;
done; (* TODO: look for theorem in isabelle *)


lemma finite_non_singleton: "\<lbrakk>finite B;a\<in>B;B\<noteq>{a}\<rbrakk> \<Longrightarrow> card B >1"
apply auto;
apply (case_tac "card B")
apply auto;
apply (drule card_eq_SucD);
apply auto;
done; 

lemma non_singleton_rev: "\<lbrakk>card B >1\<rbrakk>\<Longrightarrow>\<exists> a b. a\<in>B\<and>b\<in>B\<and> a\<noteq>b"
apply (case_tac "card B")
apply auto;
apply (case_tac nat);
apply auto;
apply (drule card_eq_SucD);
apply clarify;
apply (drule card_eq_SucD);
apply auto;
done;


lemma boundedSet_pre: "\<forall> A. finite A \<longrightarrow>card A=n\<longrightarrow>A\<noteq>{}\<longrightarrow>  (\<exists> y\<in>A. \<forall> x \<in> A. ((f::'a\<Rightarrow>nat) x \<le> f y))"
apply (induct_tac n);
apply simp+;
apply clarify;
apply (drule card_eq_SucD);
apply clarify;
apply (drule_tac x=B in spec);
apply auto;
apply (case_tac "f b>f y");
apply auto;
done;

lemma boundedSet: "\<lbrakk>finite A; A\<noteq>{}\<rbrakk>\<Longrightarrow>\<exists> y\<in>A. \<forall> x \<in> A. ((f::'a\<Rightarrow>nat) x \<le> f y)"
by (insert boundedSet_pre, auto);
section{*set and lists *}

lemma length_and_card[rule_format]: " \<forall> B. set xs \<subseteq> B \<longrightarrow> distinct xs \<longrightarrow>  finite B \<longrightarrow> length xs \<le> card B"
apply (induct_tac xs);
apply auto;
apply (case_tac "B={a}");
apply simp+;
apply (subgoal_tac "set list={}");
apply force;
apply (drule set_exclude);
apply blast;
apply force;
apply (drule_tac x="B-{a}" in spec);
apply (subgoal_tac "card B >1")
apply force;
apply (rule finite_non_singleton);
apply auto;
done;


lemma insert_union: " \<lbrakk>b\<notin>B;  finite B \<rbrakk>\<Longrightarrow>({xs. distinct xs \<and> set xs \<subseteq>insert b B} = Union ({{(list_insert xs n b)|xs. distinct xs \<and> set xs \<subseteq>B}|n. n\<le>card B}) \<union>{xs. distinct xs \<and> set xs \<subseteq>B})"
apply auto;
apply (case_tac "xa=b");
apply auto;
apply (frule set_list_insert)
apply auto;
apply (drule_tac x="{list_insert ys n b|ys. distinct ys \<and> set ys \<subseteq> B}" in spec)
apply auto;
apply (drule_tac x=n in spec);
apply auto;
apply (subgoal_tac "set ys \<subseteq> B")
apply (rotate_tac -1, drule length_and_card)
apply auto;
done;

lemma finite_distinct_list_pre[rule_format]: " \<forall> A. (finite A \<longrightarrow>card A=n \<longrightarrow>finite {xs. distinct xs \<and> set xs \<subseteq>A})"
apply (induct_tac n)
apply auto;
apply (drule Finite_Set.card_eq_SucD)
apply clarify;
apply (drule_tac x=B in spec);
apply auto;
apply (drule insert_union);
apply auto;
done;


lemma finite_distinct_list[rule_format]: "  \<lbrakk>finite A\<rbrakk>\<Longrightarrow>finite {xs. distinct xs \<and> set xs \<subseteq>A}"
apply (insert finite_distinct_list_pre);
apply auto;
done;

end;